#ifndef MC_H
#define MC_H
#include "CuTe.h"
#include "libCuTe/sm_control.h"


//#define OLD_WEIGHT_SUM
#define TEST_INTEGRAND
#define ALLCUBA
//#define ALLCUBAV

namespace CuTe {

namespace generation {

using std::ofstream;
using NS_CS_Params::CS_Parameter;
using namespace NS_CUBA_integration;
using namespace NS_QDCS;

using namespace Shared_Memory_Control;

using namespace CuTe::NS_Utils::NS_pow;
using namespace CuTe::NS_Utils::NS_min_max;


typedef double(*PS_Func)(const vector<double>&);

enum e_PSCoords {e_PSCoords,e_k_T=e_PSCoords, e_eta, e_kb_T, e_etab, e_M2
#ifdef ALLCUBA
	, e_z0, e_z1
#ifdef ALLCUBAV
	, e_xT
#endif
#endif
	, e_s, e_PSCoords_N};
static const int dim
#ifndef ALLCUBA
	=4;
#else
#ifdef ALLCUBAV
	=7;
#else
	=6;
#endif
#endif


namespace PS_Functions {

static inline double test_u(const vector<double>& coords){
	return 0.5*(coords[0]+coords[1]);
}
static inline double test_v(const vector<double>& coords){
	return 0.5*(coords[0]-coords[1]);
}
static inline double test_x(const vector<double>& coords){
	return coords[0];
}
static inline double test_y(const vector<double>& coords){
	return coords[1];
}

static inline double test_f(const vector<double>& coords){
	//return pow_2(coords[0]-10.*coords[1]);
	//return pow_3(coords[0]+coords[1]);
	return M_PI*sin(M_PI_2*coords[0]);
}

static inline double D(const vector<double>& coords){
	return cosh(coords[e_eta]-coords[e_etab])-coords[e_M2]/(2.*coords[e_k_T]*coords[e_kb_T]);
}

static inline double D2(const vector<double>& coords){
	return pow_2(D(coords));
}

static inline double dv(const vector<double>& coords){
	return 1./(4.*pow_4(2.*PI)*sqrt(1.-D2(coords)));
}

static inline double q0(const vector<double>& coords){
	return coords[e_k_T]*cosh(coords[e_eta])+coords[e_kb_T]*cosh(coords[e_etab]);
}

static inline double mperp2(const vector<double>& coords){
	return pow_2(coords[e_k_T])
			+2.*coords[e_k_T]*coords[e_kb_T]*cosh(coords[e_eta]-coords[e_etab])
			+pow_2(coords[e_kb_T]);
}

static inline double y(const vector<double>& coords){
	double k=coords[e_kb_T]/coords[e_k_T];
	double E=exp(coords[e_etab]-coords[e_eta]);
	double y=(1.+k*E)/(1.+k/E);
	return coords[e_eta]+0.5*log(y);
}

static inline double phistar(const vector<double>& coords){
	double _phistar=tan(0.5*acos(D(coords)));
	_phistar*=cosh(0.5*(coords[e_eta]-coords[e_etab]));
	return 1./_phistar;
}

static inline double ln_phistar(const vector<double>& coords){
	double _phistar=tan(0.5*acos(D(coords)));
	_phistar*=cosh(0.5*(coords[e_eta]-coords[e_etab]));
	return -log(_phistar);
}

static inline double qT2(const vector<double>& coords){
	return pow_2(coords[e_k_T])
			+2.*coords[e_k_T]*coords[e_kb_T]*D(coords)
			+pow_2(coords[e_kb_T]);
}

static inline double qT(const vector<double>& coords){
	return sqrt(qT2(coords));
}

static inline double q0max(const vector<double>& coords){
	return (coords[e_s]+coords[e_M2])/(2.*sqrt(coords[e_s]));
}

static inline double Vreal(double M2, double q0max){
	double tmp= sqrt(pow_2(q0max)-M2);
	return 	(
				q0max*tmp - 0.5*M2*log(pow_2(q0max+tmp)/M2)
			)/(
				8.*pow_3(2.*M_PI)
			);
}

static inline double Vreal(const vector<double>& coords){
	return Vreal(coords[e_M2],q0max(coords));
}

}  // namespace PS_Functions


class Differential_Distribution{
public:
//	string name;
typedef double val_T;
typedef long long int cnt_T;
protected:
	//Distribution of N bins in [min,max]
	double min;
	double max;
	int N;

	double dx;	//	= (max-min)/N
	double xd;	//	= 1./dx

	int ncores;
	bool attached;
	mutable Mutex mutex;
	SM_Variable<int> SM_merge_count;

	//Local Variables for every bin i
	//addval(..) -> Sum up every value i for one iteration a
	//merge_me() resets them
	vector<cnt_T> nai;
	vector<val_T> wvai;
	vector<val_T> wfai;
	vector<val_T> Svai;
	vector<val_T> Sfai;

	//Shared variables
	//merge_me() -> Sum of the results above over every i and core
	//merge_all() resets them
	SM_Variable<cnt_T> SM_na;

	//Shared variables
	//merge_me() -> Sum of the results above over every core
	//merge_all() resets them
	SM_Array<cnt_T> SM_nai;
	SM_Array<val_T> SM_wvai;
	SM_Array<val_T> SM_wfai;
	SM_Array<val_T> SM_Svai;
	SM_Array<val_T> SM_Sfai;

	//Shared variables
	//merge_all() -> Sum of the results above over every iteration a
	SM_Array<cnt_T> SM_ni;
	SM_Array<val_T> SM_Wvi;
	SM_Array<val_T> SM_Vi;
	SM_Array<val_T> SM_Wfi;
	SM_Array<val_T> SM_Fi;

	//Shared variables
	//merge_all() -> Sum of the results above over every i and a
	SM_Variable<cnt_T> SM_n;


public:

	inline int merge_me(int SID,int iteration){
		NS_Utils::DOUT dout("Differential_Distribution::merge_me","SID,iteration");
		dout()	<<"SID="<<SID	<<"\t"<<"iteration="<<iteration	<<endl;

		vector<cnt_T>::iterator	ina		= nai.begin();
		vector<val_T>::iterator	iwva	= wvai.begin();
		vector<val_T>::iterator	iSva 	= Svai.begin();
		vector<val_T>::iterator	iwfa	= wfai.begin();
		vector<val_T>::iterator	iSfa	= Sfai.begin();
		cnt_T* iSM_na		= SM_nai.begin();
		val_T* iSM_wva		= SM_wvai.begin();
		val_T* iSM_Sva		= SM_Svai.begin();
		val_T* iSM_wfa		= SM_wfai.begin();
		val_T* iSM_Sfa		= SM_Sfai.begin();
		for(int i=0;i<N;++i){
			if(*ina){
				(*iSM_wva) +=(*iwva);
				(*iSM_Sva) +=(*iSva);
				(*iSM_wfa) +=(*iwfa);
				(*iSM_Sfa) +=(*iSfa);
				if(*iSM_na){
					val_T n		= static_cast<val_T>( (*iSM_na)+(*ina) );
					val_T tmp	= static_cast<val_T>(*ina)/static_cast<val_T>(n*(*iSM_na));
							n  /= static_cast<val_T>(*ina);
					(*iSM_Sva)	+= tmp*pow_2(n*(*iwva)-(*iSM_wva));
					(*iSM_Sfa)	+= tmp*pow_2(n*(*iwfa)-(*iSM_wfa));
				}
				(*iSM_na) +=(*ina);

				(*SM_na)  +=(*ina);

			}
			(*(ina++))=0.;
			(*(iwva++))=0.;
			(*(iSva++))=0.;
			(*(iwfa++))=0.;
			(*(iSfa++))=0.;
			++iSM_na;
			++iSM_wva;
			++iSM_Sva;
			++iSM_wfa;
			++iSM_Sfa;
		}

		return (++(*SM_merge_count));
	}

	inline int merge_all(int SID,int iteration){
		NS_Utils::DOUT dout("Differential_Distribution::merge_all","SID,iteration");
		dout()	<<"SID="<<SID	<<"\t"<<"iteration="<<iteration	<<endl;

		int _tmp = (*SM_merge_count)%ncores;
		if(_tmp) return _tmp;	// Not ready for merge_all

		if((*SM_merge_count)==0) return 0;	// already merged

		val_T n = static_cast<val_T>(*SM_na);
		(*SM_n)	+=(*SM_na);
		(*SM_na) = 0;


		cnt_T* ina		= SM_nai.begin();
		val_T* iwva		= SM_wvai.begin();
		val_T* iSva		= SM_Svai.begin();
		val_T* iwfa		= SM_wfai.begin();
		val_T* iSfa		= SM_Sfai.begin();
		cnt_T* in		= SM_ni.begin();
		val_T* iWv		= SM_Wvi.begin();
		val_T* iV		= SM_Vi.begin();
		val_T* iWf		= SM_Wfi.begin();
		val_T* iF		= SM_Fi.begin();
		for(int i=0;i<N;++i){
			if(*ina){
				val_T	na	= static_cast<val_T>(*ina);
				val_T tmpWv = 1.;
				val_T tmpWf = 1.;
				if(*ina>1){
					(*in) 	   += (*ina);
					val_T tmp2	= na*(n-1.);

					tmpWv = tmp2/( n*na*(*iSva) + (n-na)*pow_2(*iwva) );
					tmpWf = tmp2/( n*na*(*iSfa) + (n-na)*pow_2(*iwfa) );
#ifndef OLD_WEIGHT_SUM
					tmpWv*=pow_2(*iwva);
					tmpWf*=pow_2(*iwfa);
#endif
#ifdef OLD_WEIGHT_SUM
					(*iWv) += tmpWv;
					(*iV)  += tmpWv*(*iwva);
					(*iWf) += tmpWf;
					(*iF)  += tmpWf*(*iwfa);
#endif
				}
#ifdef OLD_WEIGHT_SUM
				else{
					tmpWv /= fabs(*iwva);
					tmpWf /= fabs(*iwfa);

					(*iV)  += tmpWv;
					(*iF)  += tmpWf;

					(*iWv) += pow_2(tmpWv);
					(*iWf) += pow_2(tmpWf);
				}
#endif
#ifndef OLD_WEIGHT_SUM
				(*iWv) += tmpWv;
				(*iV)  += tmpWv*(*iwva);
				(*iWf) += tmpWf;
				(*iF)  += tmpWf*(*iwfa);
#endif
			}
			(*(ina++))=0.;
			(*(iwva++))=0.;
			(*(iSva++))=0.;
			(*(iwfa++))=0.;
			(*(iSfa++))=0.;
			++in;
			++iWv;
			++iV;
			++iWf;
			++iF;
		}


		(*SM_merge_count)=0;
		return 0;
	}

	inline int get_i(double x)const{
		return static_cast<int>( (x-min)*xd );
	}

	inline void add_val(double x, double fx, double wx=1., double vx=1.){
		if(!attached) return;

		int i = get_i(x);
		if(i<0 || i>=N) return;

		cnt_T* n = &(nai[i]);
		wvai[i]	+= (wx*=vx);
		wfai[i] += (wx*fx);
		if((*n)++){
			val_T tmp = 1./static_cast<val_T>((*n)*((*n)-1.));
			Svai[i]	+= tmp*pow_2(wvai[i]-(wx*=static_cast<val_T>(*n)));
			Sfai[i]	+= tmp*pow_2(wfai[i]-wx*fx);
		}
	}

	inline void to_File(const string Filename){
		NS_Utils::DOUT dout("Differential_Distribution",Filename);
		if(!attached) return;

		ofstream of;
		of.open(Filename.c_str());

		of	<< "#"
			<< "x\t"
			<< "f(x)\t"
			<< "dx\t"
			<< "df\t"
			<< "V\t"
			<< "dV\t"
			<< "n\t"
		    << endl;

		double x = min+0.5*dx;

		cnt_T* in		= SM_ni.begin();
		val_T* iWv		= SM_Wvi.begin();
		val_T* iV		= SM_Vi.begin();
		val_T* iWf		= SM_Wfi.begin();
		val_T* iF		= SM_Fi.begin();
		val_T Sv =0.;
		val_T  V =0.;
		val_T Sf =0.;
		val_T  F =0.;
		for(int i=0; i<N;++i){
			if(*in){
#ifndef OLD_WEIGHT_SUM
				(*iWv)=1./(*iWv);
				(*iWf)=1./(*iWf);

				 V += ( (*iV)*=(*iWv) );
				 F += ( (*iF)*=(*iWf) );

				Sv += ( (*iWv)*=pow_2(*iV) );
				Sf += ( (*iWf)*=pow_2(*iF) );

				of	<< x << "\t"
					<< (*iF)*xd << "\t"
					<< 0.5*dx << "\t"
					<< sqrt(*iWf)*xd << "\t"
					<< (*iV)*xd << "\t"
					<< sqrt(*iWv)*xd << "\t"
					<< *in << "\t"
				    << endl;
#endif
#ifdef OLD_WEIGHT_SUM
				(*iWv)=1./(*iWv);
				(*iWf)=1./(*iWf);

				Sv += (*iWv);
				 V += (*iWv)*(*iV);
				Sf += (*iWf);
				 F += (*iWf)*(*iF);

				of	<< x << "\t"
					<< (*iWf)*(*iF)*xd << "\t"
					<< 0.5*dx << "\t"
					<< sqrt(*iWf)*xd << "\t"
					<< (*iWv)*(*iV)*xd << "\t"
					<< sqrt(*iWv)*xd << "\t"
					<< *in << "\t"
				    << endl;
#endif
			}
			x+=dx;
			++in;
			++iWv;
			++iWf;
			++iV;
			++iF;
		}
		of.close();
		dout()	<< Filename <<"\tn="<<*SM_n	<<endl;//<<"\tnc="<<*SM_n_contribute
		dout()	<<"SV="<<V <<" +- "<<sqrt(Sv)	<<endl;
		dout()	<<"SF="<<F <<" +- "<<sqrt(Sf)	<<endl;
	}

	inline void reinit(){
		NS_Utils::DOUT dout("Differential_Distribution::reinit");
		(*SM_merge_count)=0;

		 nai.assign(N,0.);
		Svai.assign(N,0.);
		wvai.assign(N,0.);
		wfai.assign(N,0.);
		Sfai.assign(N,0.);

		(*SM_na)=0.;

		 SM_nai.init(N,0.);
		SM_Svai.init(N,0.);
		SM_wvai.init(N,0.);
		SM_wfai.init(N,0.);
		SM_Sfai.init(N,0.);

		 SM_ni.init(N,0.);
		SM_Wvi.init(N,0.);
		SM_Wfi.init(N,0.);
		 SM_Vi.init(N,0.);
		 SM_Fi.init(N,0.);

		 (*SM_n)=0;

		attached = true;
	}

	inline void init(double min, double max, int N, int ncores){
		NS_Utils::DOUT dout("Differential_Distribution::init","min,max,N,ncores");
		dout()	<<"min="<<min	<<"\t"<<"max="<<max	<<"\t"<<"N="<<N	<<"\t"<<"ncores="<<ncores	<<endl;

		if(N){
			this->min	= min;
			this->max	= max;
			this->N		= N;
			this->ncores= ncores;
			dx = (max-min)/static_cast<double>(N);
			xd = 1./dx;
			reinit();
		}else{
			attached=false;
		}
	}

//protected:
	inline Differential_Distribution()
//	:name()
	:min(0.),max(0.),N(0)
	,dx(0.),xd(0.)
	,ncores(-1)
	,attached(false)
	{
		NS_Utils::DOUT dout("Differential_Distribution");
		SM_n.status("SM_n");
	}
//public:
//	inline Differential_Distribution(const string& name)
//	:name(name)
//	,min(0.),max(0.),N(0)
//	,dx(0.),xd(0.)
//	,ncores(-1)
//	,attached(false)
//	{
//		NS_Utils::DOUT dout("Differential_Distribution",name);
//		SM_n.status("SM_n");
//	}


	inline ~Differential_Distribution(){
		NS_Utils::DOUT dout("~Differential_Distribution");
	}

	inline void lock(){return mutex.lock();}
	inline bool try_lock(){return mutex.try_lock();}
	inline void unlock(){return mutex.unlock();}
};


class Differential_Distribution_Set{
protected:
	int ncores;
	const int ndists;
	enum e_dists {	e_dists_begin, e_qT = e_dists_begin, e_qT2, e_y, e_phistar, e_lnphistar
#ifdef TEST_INTEGRAND
					,e_test_x , e_test_y, e_test_u, e_test_v
#endif
					,e_dists_N};
	vector<Differential_Distribution> dists;
	vector<string> dist_names;
	vector<bool> dist_eval;
	vector<PS_Func> dist_funcs;

	int iteration;
public:

	inline void init(const SOPS_CuTe &OP){
		cout	<< "-> Differential_Distribution_Set::init(config,fail)" << endl;

		ncores = OP.CUBA.CUBA_cores();

		iteration=1;
		int N;
		double min,max;
		dist_eval.assign(ndists,false);
#ifdef TEST_INTEGRAND
		N=OP.Plot.qT_N();
		if(N){
			min = 0.;
			max = 1.;
			dists[e_test_u].init(min,max,N,ncores);
			dist_names[e_test_u] = "e_test_u";
			dist_eval[e_test_u]=true;
			dist_funcs[e_test_u]=&PS_Functions::test_u;
		}
		if(N){
			min = -0.5;
			max = 0.5;
			dists[e_test_v].init(min,max,N,ncores);
			dist_names[e_test_v] = "e_test_v";
			dist_eval[e_test_v]=true;
			dist_funcs[e_test_v]=&PS_Functions::test_v;
		}
		if(N){
			min = 0.;
			max = 1.;
			dists[e_test_x].init(min,max,N,ncores);
			dist_names[e_test_x] = "e_test_x";
			dist_eval[e_test_x]=true;
			dist_funcs[e_test_x]=&PS_Functions::test_x;
		}
		if(N){
			min = 0.;
			max = 1.;
			dists[e_test_y].init(min,max,N,ncores);
			dist_names[e_test_y] = "e_test_y";
			dist_eval[e_test_y]=true;
			dist_funcs[e_test_y]=&PS_Functions::test_y;
		}
#else

		double s = pow_2(op.sqrts());
		double M = op.M();
		double gen = op.accuracy();
		double q0max = (s+pow_2(M))/(2.*sqrt(s));
		double y_max = acosh(q0max/M);
		double qT_max = sqrt(pow_2(q0max)-pow_2(M));

		N=op.qT_N();
		if(N){
//			min = op.qT_min();	//FIXME
//			max = op.qT_max();
			dists[e_qT].init(min,max,N,ncores);
			dist_names[e_qT] = "qT";
			dist_eval[e_qT]=true;
			dist_funcs[e_qT]=&PS_Functions::qT;

			min*=min;
			max*=max;
			dists[e_qT2].init(min,max,N,ncores);
			dist_names[e_qT2] = "qT2";
			dist_eval[e_qT2]=true;
			dist_funcs[e_qT2]=&PS_Functions::qT2;
		}

//		N=config.read<int>("y_N",0);
//		if(N){
//			min = gsl_max(config.read<double>("y_min",-y_max),-y_max);
//			max = gsl_min(config.read<double>("y_max",+y_max),+y_max);
//			dists[e_y].init(min,max,N,ncores);
//			dist_names[e_y] = "y";
//			dist_eval[e_y]=true;
//			dist_funcs[e_y]=&PS_Functions::y;
//		}
//
//		N=config.read<int>("phistar_N",0);
//		if(N){
//			min = gsl_max(config.read<double>("phistar_min",0.),0.);
//			max = gsl_min(config.read<double>("phistar_max",M_PI_2),2.*M_PI);
//			dists[e_phistar].init(min,max,N,ncores);
//			dist_names[e_phistar] = "phistar";
//			dist_eval[e_phistar]=true;
//			dist_funcs[e_phistar]=&PS_Functions::phistar;
//
//			if(min<gen) min=gen;
//			min = log(min);
//			max = log(max);
//			dists[e_lnphistar].init(min,max,N,ncores);
//			dist_names[e_lnphistar] = "lnphistar";
//			dist_eval[e_lnphistar]=true;
//			dist_funcs[e_lnphistar]=&PS_Functions::ln_phistar;
//
//		}
#endif
		cout	<< "<- Differential_Distribution_Set::init(config,fail)" << endl;
	}

	inline Differential_Distribution_Set()
	:ncores(-1)
	,ndists(e_dists_N)
	,dists(vector<Differential_Distribution>(ndists))
	,dist_names(vector<string>(ndists))
	,dist_eval(vector<bool>(ndists))
	,dist_funcs(vector<PS_Func>(ndists))
	,iteration(1){
		cout << "Construct Differential_Distribution_Set" << endl;
	}

	inline Differential_Distribution_Set(const SOPS_CuTe& OP)
	:ncores(-1)
	,ndists(e_dists_N)
	,dists(vector<Differential_Distribution>(ndists))
	,dist_names(vector<string>(ndists))
	,dist_eval(vector<bool>(ndists))
	,dist_funcs(vector<PS_Func>(ndists))
	,iteration(1){
		cout << "Construct Differential_Distribution_Set(config)"
			 << "\tndists = " << static_cast<int>(e_dists_N)
			 << endl;
		init(OP);
	}

	inline void merge(int SID){
		int start = SID-1;

//		cout << "["<<SID<<"]("<<iteration<<")->merge" << endl;
		// List of dists i_have_to_merge
		vector<bool> i_have_to_merge(ndists,false);
		// # of dists i_have_to_merge
		int i_have_still_to_merge= 0;
		for(int i=0; i<ndists; ++i){
			if(dist_eval[i]){	// Find out if a dist is to merge
				i_have_to_merge[i]=true;
				i_have_still_to_merge++;
			}
		}

//		if(i_have_still_to_merge==1){
//			for(int i=0; i<ndists; ++i){
//				if(i_have_to_merge[i]){
//					dists[i].lock();
//					dists[i].merge_me(SID,iteration);
//					dists[i].merge_all(SID,iteration);
//					dists[i].unlock();
//					i_have_to_merge[i]=false;
//					i_have_still_to_merge--;
//				}
//			}
//			cout << "["<<SID<<"]("<<iteration<<")<-merge" << endl;
//			return;
//		}

		// Copy for the second step
		vector<bool> we_have_to_merge(i_have_to_merge);
		int we_have_still_to_merge = i_have_still_to_merge;


		int wcount=0;
		while(i_have_still_to_merge>1){
//			cout << "["<<SID<<"]("<<iteration<<")["<<wcount<<"] i_have_still_to_merge="<< i_have_still_to_merge << endl;
			for(int i=0; i<ndists; ++i){
				if(i_have_to_merge[i] && dists[i].try_lock()){
					dists[i].merge_me(SID,iteration);
					dists[i].unlock();
					i_have_to_merge[i]=false;
					i_have_still_to_merge--;
					wcount=0;
				}
			}
			if(++wcount>ndists) break;
		}
		for(int i=0; i<ndists; ++i){
			if(i_have_to_merge[i]){
				dists[i].lock();
				dists[i].merge_me(SID,iteration);
				dists[i].unlock();
//				i_have_to_merge[i]=false;
//				i_have_still_to_merge--;
			}
		}


		wcount=0;
		while(we_have_still_to_merge>1){
//			cout << "["<<SID<<"]("<<iteration<<")["<<wcount<<"] we_have_still_to_merge="<< we_have_still_to_merge << endl;
			for(int i=0; i<ndists; ++i){
				if(we_have_to_merge[i] && dists[i].try_lock()){
					int tmp = dists[i].merge_all(SID,iteration);
					dists[i].unlock();
					if(tmp==0){
						we_have_to_merge[i]=false;
						we_have_still_to_merge--;
						wcount=0;
					}
				}
			}
			if(++wcount>ndists) break;
		}
//		while(we_have_still_to_merge){
			for(int i=0; i<ndists; ++i){
				if(we_have_to_merge[i]){
					dists[i].lock();
//					int tmp =
							dists[i].merge_all(SID,iteration);
					dists[i].unlock();
//					if(tmp==0){
//						we_have_to_merge[i]=false;
//						we_have_still_to_merge--;
//					}
				}
			}
//		}
//		cout << "["<<SID<<"]("<<iteration<<")<-merge" << endl;
	}

	inline void add_value(int SID, const vector<double>& coords, int iter, double fx, double wx=1., double vx=1.){
		if(iter!=iteration){
			merge(SID);
			iteration=iter;
		}

		for(int i=0; i<ndists; ++i){
			if(dist_eval[i]){
				dists[i].add_val(dist_funcs[i](coords),fx,wx,vx);
			}
		}
	}

	inline void to_File(const string& filename){
		for(int i=0; i<ndists; ++i){
			if(dist_eval[i]){
				string tmp = filename + string("_") + dist_names[i] + string(".out");
				dists[i].to_File(tmp);
			}
		}
	}

};

inline void test_DDS(){
	string filename("MCtest.in");
	SOPS_CuTe OP(filename);
	Cuba_Params cuba;
	cuba.init(OP.CUBA,dim,2);
}

class MC_params{
protected:
	//init(op)
	Cuba_Params cuba_params;
	Quad_Differential_Crosssection QDCS;
	Differential_Distribution_Set dists;
	double gen;
	int seed;
//	double kTmin,etamax;
	vector<double> coords;

	//init()
	double q0max;
	double qTmax2;
	double Vreal;
	double mp_max;

	//update()
	double cuba_dv;

public:
	SM_Serial_ID SID;
protected:

	inline void init(double qTmax){
		cout << "\t->MC_Params::init()"<<endl;
		q0max=PS_Functions::q0max(coords);

		qTmax2 = pow_2(q0max)-coords[e_M2];
		Vreal = PS_Functions::Vreal(coords);
		if(qTmax > 0. && (qTmax*=qTmax) < qTmax2){
			qTmax2=qTmax;
			Vreal-=PS_Functions::Vreal(coords[e_M2]+qTmax2,q0max);
		}
		mp_max = sqrt(coords[e_M2]+qTmax2);
#ifdef TEST_INTEGRAND
		cuba_params.set(NS_CUBA_integration::CUBA_R_VEGAS,dim,2);
#else
		cuba_params.set(NS_CUBA_integration::CUBA_R_VEGAS,dim,1);
#endif

		cout << "\t<- MC_Params::set(qTmax,kTmin,etamax)"<<endl;
	}

	inline void test_set_coords(const double xx[]){
		for(int i=0; i<dim; ++i) coords[i]=xx[i];
		cuba_dv=1.;
	}

	inline void set_coords(const double xx[]){
		//		cout <<"->MC_CUBA::set_coords()"<<endl;
				double min,max,tmp,u,u2,v,iv,etap,Cp,D;
				cuba_dv=4.;

				// u
				min=0.;//p.M;
				max=mp_max-min;
				u=min + xx[e_k_T]*max;
				u2=pow_2(u);
				cuba_dv*=max;

				// v
				min=max_of(0.,u2-qTmax2);//M2;
				max=u2-min;
				if(max<=0.) cout << "***** Error v"<<endl;
				v=min + xx[e_kb_T]*max;
				iv=1./v;
				cuba_dv*=max;

				// kT, kbT
				tmp=sqrt(u2-v);
				coords[e_k_T ]= 0.5*(u + tmp);
				coords[e_kb_T]= 0.5*(u - tmp);
				cuba_dv/=(4.*tmp);

			//	D
				min = max_of( 1.-2.*coords[e_M2]*iv , -1.);
				max = min_of( 1. , 1.+ 2.*(qTmax2-u2)*iv ) - min;
				D=min + xx[e_etab]*max;
				Cp = D + 2.*coords[e_M2]*iv;
				etap=acosh(Cp);
				cuba_dv*=max/sqrt(Cp*Cp-1.);

				tmp=sqrt(u2+0.5*v*(Cp-1.));
				max=acosh(q0max/tmp);
				min=-max-0.5*log( (coords[e_k_T]+coords[e_kb_T]*exp(-etap))/(coords[e_k_T]+coords[e_kb_T]*exp(+etap)) );
				max*=2.;
				coords[e_eta]=min + xx[e_eta]*max;
				cuba_dv*=max;

				coords[e_etab]=coords[e_eta]-etap;

				if(Discard()) return smooth_edge(xx);

#ifdef ALLCUBA
//				coords[e_z0] = (1.-p.xi[0])*xx[e_z0]+p.xi[0];
//				coords[e_z1] = (1.-p.xi[1])*xx[e_z1]+p.xi[1];
				coords[e_z0] = xx[e_z0];
				coords[e_z1] = xx[e_z1];
#ifdef ALLCUBAV
				coords[e_xT] = xx[e_xT];
#endif
//				cuba_dv*=(1.-p.xi[0])*(1.-p.xi[1]);
#endif
//				cuba_dv *= PS_Functions::dv(coords);
				cuba_dv/=(4.*pow_4(2.*PI)*sqrt(1.-pow_2(D)));

	}

	inline bool Discard(){
		double x=PS_Functions::q0(coords);
		if(x>q0max)	return true;
		x=PS_Functions::D2(coords);
		if(x>=1.) return true;
		return false;
	}

	inline int array_find_min(const double xx[], int n)const{
		int min=0;
		for(int i=1; i<n; ++i){
			if(xx[i]<xx[min]) min=i;
		}
		return min;
	}

	inline void smooth_edge(const double xx[]){
		double xx_prime[2*dim];
		for(int i=0; i< dim; ++i){
			xx_prime[i]= xx[i];
			xx_prime[dim+i]= 1.-xx[i];
		}
		int min = array_find_min(xx_prime,2*dim);
		if(min<dim) xx_prime[min]*=2.;
		else xx_prime[min-dim] = 1.-2.*xx_prime[min];
		return set_coords(xx_prime);
	}


	inline void init(const SOPS_CuTe& OP){
//		cout << "->MC_Params::init(config,fail_string)"<<endl;
		cuba_params.init(OP.CUBA,dim,/*p.order+*/1);
//		QDCS.init(op);FIXME
		dists.init(OP);

		gen			= OP.CUBA.accuracy();
		seed		= OP.CUBA.CUBA_seed();

//		kTmin		= -1.;//FIXME config.read<double>("kTmin",-1.);
//		etamax		= -1.;//FIXME config.read<double>("etamax",-1.);

		coords		= vector<double>(e_PSCoords_N,0.);
		coords[e_s]	= pow_2(OP.Process.sqrts());
		coords[e_M2]= pow_2(OP.Process.M());


//		init(op.qT_max());//FIXME
//		cout 	<<"<-MC_Params::init(config,fail_string) = "<<endl;
	}

	inline void integrand(const double xx[], double ff[], double wx, int iter){
		double erg=1.;
#ifdef TEST_INTEGRAND
		test_set_coords(xx);
		erg = PS_Functions::test_f(coords);
		dists.add_value(*SID,coords,iter,erg,wx,cuba_dv);
		ff[0]=erg*cuba_dv;
		ff[1]=cuba_dv;
#else
		set_coords(xx);

		double qT	= PS_Functions::qT(coords);
		double y	= PS_Functions::y(coords);

//		erg *= exp(-PS_Functions::qT2(coords)/40.)/(2.*p.y_max);
#ifndef ALLCUBA
		CS.update(p);
#else
//		erg*=(1.-p.xi[0])*(1.-p.xi[1]);
//		CS.update_xxy<RES_LL>(&(coords[e_z0]));	//FIXME
//		erg*=QDCS.of_qT_y_zz(qT,y,&(coords[e_z0]));FIXME
#endif
		erg*=pow_3(4.*M_PI)/qT;
		dists.add_value(*SID,coords,iter,erg,wx,cuba_dv);
		erg*=cuba_dv;
//		for(int i=0; i<=p.order; ++i) ff[i]=erg*CS[i];
		ff[0]=erg;
#endif
//		cout << "CS = " << CS << endl;
	}

	static inline void CUBA_init_func(void* params, int &core){
		NS_Utils::DOUT dout("MC_params::CUBA_init_func");

		MC_params* p = static_cast<MC_params*>(params);
		p->SID.get_new_SID();

		dout()	<< &(p->SID)	<<"\t"<<"core="<<core	<<endl;
	}

	static inline void CUBA_exit_func(void* params, int &core){
		NS_Utils::DOUT dout("MC_params::CUBA_exit_func");

		MC_params* p = static_cast<MC_params*>(params);
		p->dists.merge(*(p->SID));

		dout()	<< &(p->SID)	<<"\t"<<"core="<<core	<<endl;
	}

	inline static int Integrandw(
			const int *ndim, const double xx[]
			,const int *ncomp, double ff[], void *userdata
			,const int *nvec, const int* core
			,const double* weight, const int* iter) {
		MC_params* MC = static_cast<MC_params*>(userdata);
//		cout	<<"Integrandw"	<<endl
//				<<"\t"<<"ndim="<<*ndim	<<"\t"<<"xx=";
//		for(int i=0; i<*ndim; ++i) cout	<<xx[i]<<"\t";
//		cout	<<endl
//				<<"\t"<<"ncomp="<<*ncomp	<<"\t"<<"ff="<<ff	<<"\t"<<"userdata="<<userdata	<<endl
//				<<"\t"<<"nvec="<<*nvec	<<endl
//				<<"\t"<<"weight="<<*weight	<<"\t"<<"iter="<<*iter	<<endl;
		MC->integrand(xx,ff,(*weight),(*iter));
		return 0;
	}

//	inline static int Integrandw(
//			const int *ndim, const double xx[]
//			,const int *ncomp, double ff[], void *userdata) {
//		MC_params* MC = static_cast<MC_params*>(userdata);
//		MC->integrand(xx,ff,NULL,NULL);
//		return 0;
//	}

public:

	inline MC_params()
	:cuba_params(),QDCS(),dists()
	,gen(0.)
	,seed(0)
	,coords(vector<double>(e_PSCoords_N,0.))
	,q0max(NAN)
	,qTmax2(0)
	,Vreal(NAN)
	,mp_max(-1.)
	,cuba_dv(0.)
	{
		cout << "Construct MC_params()" << endl;
	}

	inline MC_params(const SOPS_CuTe& OP)
	:cuba_params(OP.CUBA,dim,1),QDCS(),dists(OP)
	,gen(OP.CUBA.accuracy())
	,seed(OP.CUBA.CUBA_seed())
	,coords(vector<double>(e_PSCoords_N,0.))
	,q0max(NAN)
	,qTmax2(0)
	,Vreal(NAN)
	,mp_max(-1.)
	,cuba_dv(0.)
	{
		cout << "Construct MC_params(op)" << endl;
//		QDCS.init(op);FIXME
		coords[e_s]	= pow_2(OP.Process.sqrts());
		coords[e_M2]= pow_2(OP.Process.M());
//		init(op.qT_max());//FIXME
	}

	inline void cube(){
		cuba_params.set_in_ex(CUBA_init_func,CUBA_exit_func);
		cuba_params.cube(
				(cuba_integrand_t)Integrandw
				,this
				);
//		cuba_params.test(
//				1000000
//				,Integrandw
//				,this
//				,CUBA_init_func
//				,CUBA_exit_func
//				);
		dists.to_File("MC_test");
		cout	<<"VREAL="<<Vreal
//				<<"\tD_rel="<<(cuba_params.integral[1]-Vreal)*100./(Vreal)<<"%"
				<<"\tacc="<<gen*100.<<"%"
				<<endl;

	}
};


inline void generate(const SOPS_CuTe& OP){
//	cout << "-> generation::generate(string)\tCUBACORES="<<getenv("CUBACORES")<< endl;
	cout << "generation infilename = "<<OP.get_infilename() << endl;


	MC_params p(OP);

	p.cube();

//	cout << fail_string << endl;
	return;
}



}  // namespace MC_generation


}  // namespace CuTe

#endif
