#ifndef CONVOLUTIONS_H_
#define CONVOLUTIONS_H_

#include <vector>
#include <iostream>
//#include "cuba_int.h"

namespace CuTe {

namespace NS_CUBA_integration {

class Cuba_Params;

}  // namespace NS_CUBA_integration


namespace NS_Convolutions {

using std::cerr;
using std::cout;
using std::endl;
using std::vector;
using std::ostream;
using std::string;

typedef double (*CI_Function)(double z,int flavour,void* params);


class _Convolution_Interface{
public:
	static const size_t NFMAX=6;
protected:
	int nf;
	double xi;

	vector<double> Vfxi;
	double Sxi,Sxiq,Sxiqb;

	static void _update_S(const vector<double> &V, double &S, double &Sq, double &Sqb, const int nf);


	virtual void _set_nf(){
		cerr	<<"**********************************************"<<endl;
		cerr	<<"*** _Convolution_Interface::_set_nf called ***"<<endl;
		cerr	<<"**********************************************"<<endl;
	}

public:
	static ostream& _V_to_stream(ostream & os, const vector<double> &V, const int nf);

	virtual void set_nf(const size_t nf);

	_Convolution_Interface()
	:nf(),xi()
	,Vfxi(2*NFMAX+1,0.)
	,Sxi(),Sxiq(),Sxiqb(){

	}

	virtual ~_Convolution_Interface(){

	}

	inline size_t get_nf()const{ return nf;}
	inline double get_xi()const{ return xi;}
	inline double get_Sxi()const{ return Sxi; }

	const vector<double>& get_Vfxi()const{ return Vfxi; };

	double get_fxi(const int lha_flavor)const;
	double get_Sxiq(const int lha_flavor)const;
	double get_Sxiqb(const int lha_flavor)const;

	virtual ostream& to_stream(ostream &os, const int ntabs=0)const;
};

ostream& operator<<(ostream& os, const _Convolution_Interface &CI);


class _Convolution_Integrand:public _Convolution_Interface {
protected:
	_Convolution_Integrand *f;

	int dim;

	double Jac1,z1;

	double xi2,Jac2,z2;

	vector<double> Vfz;
	double Sz,Szq,Szqb;

	virtual void _update_V(vector<double> &V, const double xi, const double z, const double Jac=1){
		cerr	<<"************************************************"<<endl;
		cerr	<<"*** _Convolution_Integrand::_update_V called ***"<<endl;
		cerr	<<"************************************************"<<endl;
	}

//	virtual void _update_Vfxi(){
//		cerr	<<"***************************************************"<<endl;
//		cerr	<<"*** _Convolution_Integrand::_update_Vfxi called ***"<<endl;
//		cerr	<<"***************************************************"<<endl;
//	}
//
//	virtual void _update_Vfz(){
//		cerr	<<"**************************************************"<<endl;
//		cerr	<<"*** _Convolution_Integrand::_update_Vfz called ***"<<endl;
//		cerr	<<"**************************************************"<<endl;
//	}

	virtual void _set_nf();

	virtual void _update_xi1_xi2_z(const double xi1, const double xi2, const double xx[]);


public:

	_Convolution_Integrand& init_function(_Convolution_Integrand *f){
		this->f	= f;
		return *f;
	}

	_Convolution_Integrand& operator()(_Convolution_Integrand *f){ return init_function(f);}

	virtual void set_nf(const size_t nf);
	virtual void update_xi_z(const double xi, const double xx[]);

	_Convolution_Integrand()
	:_Convolution_Interface()
	,f(NULL)
	,dim(0)
	,Jac1(),z1()//,xi1iz1()
	,xi2(),Jac2(),z2()//,xi2iz2()
	,Vfz(2*NFMAX+1,0.)
	,Sz(),Szq(),Szqb(){

	}

	virtual	~_Convolution_Integrand(){

	}

	virtual int get_dim();
	virtual int get_dim()const;

	virtual ostream& to_stream(ostream &os, const int ntabs=0)const;

	inline double get_Sz()const{ return Sz;}
	double get_fz(const int lha_flavor)const;
	double get_Szq(const int lha_flavor)const;
	double get_Szqb(const int lha_flavor)const;
};


class Convolution_Function:public _Convolution_Integrand {
private:
	_Convolution_Integrand& init_function(_Convolution_Integrand *f);
	_Convolution_Integrand& operator()(_Convolution_Integrand *f){ return init_function(f);}
protected:
	CI_Function g;
	void *params;

	virtual void _update_xi1_xi2_z(const double xi1, const double xi2, const double xx[]);


	virtual void _update_V(vector<double> &V, const double xi, const double z, const double Jac=1);
public:

	virtual void set_nf(const size_t nf);
	virtual void update_xi_z(const double xi, const double xx[]);

	void init_function(CI_Function g, void *params=NULL);

	virtual int get_dim(){ return 0;}
	virtual int get_dim()const{ return 0;}

	Convolution_Function()
	:_Convolution_Integrand()
	,g(NULL),params(NULL){

	}

	virtual ostream& to_stream(ostream &os, const int ntabs=0)const;
};

class Convolution_Integral: public _Convolution_Interface{
protected:
	_Convolution_Integrand *f;
	NS_CUBA_integration::Cuba_Params *cuba;
	vector<double> errors;

	int _integrand(double ff[],const double xx[]);

	static int _cuba_integrand(const int *ndim, const double xx[],const int *ncomp, double ff[], void *userdata);

	virtual void _set_nf();

public:

	_Convolution_Integrand& init_Integrand(_Convolution_Integrand *f);
	_Convolution_Integrand& init_Integrand(_Convolution_Integrand *f, const NS_CUBA_integration::Cuba_Params *cuba);

	void update(const double xi);

	Convolution_Integral();
	Convolution_Integral(const Convolution_Integral&);
	Convolution_Integral& operator=(const Convolution_Integral&);
	~Convolution_Integral();

	virtual ostream& to_stream(ostream &os, const int ntabs=0)const;

};


}  // namespace NS_Convolutions
}  // namespace CuTe

#endif //CONVOLUTIONS_H_
