#ifndef TWO_P_PS_H
#define TWO_P_PS_H


#include "utils.h"
#include "Kinematics_Options.h"
#include "konst.h"
#include "gsl_integration.h"
#include "gsl_interpolation.h"

#include "Interval.h"

namespace CuTe {
namespace NS_Kinematics{

using namespace NS_Utils::NS_pow;
using namespace NS_Utils::NS_min_max;
using NS_GSL_INT::GSL_Integrator;
using NS_GSL_INT::GSL_Integrand;
using NS_GSL_Interpolation::GSL_Interpolator;

using NS_GSL_Interpolation::GSL_Interpolation_Types;
using NS_GSL_Interpolation::GSL_Interpolation_Type_Linear;

using std::cout;
using std::cerr;
using std::endl;
using std::ostream;
using std::vector;

using NS_Interval::Interval;
using NS_Interval::Variable_Limit;
using NS_Interval::Variable_Exclusion;
using NS_Interval::V_I_left;
using NS_Interval::V_I_inside;
using NS_Interval::V_I_right;

/*
 * Calculate  8*Pi * PhaseSpace-Volume of
 * 		q -> k + kb
 * 	with:
 * 		a < k_T		&&		a < kb_T
 * 		|eta| < b	&&		|etab| < b
 * 	excluding:
 * 		b0 < |eta| < b1	&& b0 < |etab| < b1
 *
 */
class Two_P_PS {
public:
	static const int intMem=500000;
	static const double acc_rel;
	static const double acc_abs;
protected:
	GSL_Integrator gsl_int;

	//Momenta
	//	q
	double M;	//= Sqrt(q^2)
	double qT;	//= Sqrt(q_1^2 + q_2^2)
	double y;	//= ArcTanh( q_3/q_0 )

	double lM2;	//= q^2/q_0^2
	double lT;	//= q_T/q_0
	double ly;	//= q_3/q_0	== tanh(y)

	//k
	Variable_Limit L_nu;
	double nu;	//=k_3/k_0
	double Snu;	//=sqrt(1-nu^2)

	Variable_Limit L_u;

	//Cuts
	bool cuts;

	// Cut kT
	double a;
	double aoE;

	// Cut eta
	double b;
	double thb;

	//Exclusion
	bool ex;
	double thb0,thb1;
	Variable_Exclusion E_nu_p;
	Variable_Exclusion E_nu_m;

	Variable_Exclusion E_u_m;
	Variable_Exclusion E_u_p;

	vector<double> integration_bins;


	//	nu-Integrand = u-Integral of upper minus lower u-limits
	inline double f(const double alpha_m, const double alpha_p)const{
		return(
					lM2*(
						(1.-ly*nu)*(
							asin(sqrt(alpha_p))
							-
							asin(sqrt(alpha_m))
						)
						-
						lT*Snu*(
							sqrt(alpha_p*(1.-alpha_p))
							-
							sqrt(alpha_m*(1.-alpha_m))
						)
					)/pow_3(sqrt(
							pow_2(1.-ly*nu)
							-
							pow_2(lT*Snu)
						))/**Jac_nu*/*I_PI
				);
	}
	inline double f(const Interval &I_alpha)const{
		return	f(I_alpha.min,I_alpha.max);
	}

	inline double _fthb(const double thb)const{
		return (thb+ly)/(thb+nu);
	}

//	inline void _calc_nu(const double x){
//		nu		= L_nu.get_value_linear(x);
//
//		if(ex){
//			if(E_nu_p.is_valid() && (E_nu_p.position_of(nu) != V_I_left))
//				nu += E_nu_p.size();
//			if(E_nu_m.is_valid() && (E_nu_m.position_of(nu) != V_I_right))
//				nu -= E_nu_m.size();
//		}
//
//		Snu		= sqrt(1.-pow_2(nu));
//	}

	inline static double u_to_alpha(const double u, const double um, const double iDu){
		return (u-um)*iDu;
	}

	inline static Interval& u_limit_to_alpha(Interval &I_u, const double um, const double iDu){
		I_u.min	= u_to_alpha(I_u.min,um,iDu);
		I_u.max	= u_to_alpha(I_u.max,um,iDu);
		return I_u;
	}

	inline bool _calc_alphas(){

		L_u.init_min_max(0.,1.);
		{
			const double num1(1.-ly*nu);
			const double num2(lT*Snu);
			const double den(0.5*lM2/(pow_2(num1)-pow_2(num2)));
			L_u.add_limit_min_max( (num1-num2)*den , (num1+num2)*den );	//u_+-

			if(L_u.is_unvalid()){
//				cerr	<<"u"<<endl;
				return false;
			}
		}

		const double um(L_u.min);
		const double iDu(1./L_u.size());

		if(a>0.){
			L_u.add_limit_min(aoE/Snu);
			L_u.add_limit_max( ( (1.-ly*nu)-sqrt( pow_2(ly-nu)+pow_2(aoE*Snu) ) )/pow_2(Snu) );

			if(L_u.is_unvalid()) return false;
		}

		if(b>0.){
			L_u.add_limit_max(_fthb(sign_of(nu-ly)*thb));

			if(L_u.is_unvalid()){
//				cerr	<<"b"<<endl;
				return false;
			}
		}

		if(ex){
			E_u_m.init(_fthb(-thb0),_fthb(-thb1));
			E_u_p.init(_fthb(+thb0),_fthb(+thb1));

			E_u_m.apply_to_limit(L_u);
			E_u_p.apply_to_limit(L_u);

			if(L_u.is_unvalid()) return false;

			if(E_u_m.is_valid()) u_limit_to_alpha(E_u_m,um,iDu);
			if(E_u_p.is_valid()) u_limit_to_alpha(E_u_p,um,iDu);
		}

		u_limit_to_alpha(L_u,um,iDu);

		return true;
	}

//	inline bool _set_nu_calc_alphas(const double x){
//		const double num1(1.-ly*nu);
//		const double num2(lT*Snu);
//		const double den(pow_2(num1)-pow_2(num2));
//
//		const double um(0.5*lM2*(num1-num2)/den);
//		const double iDu(den/(lM2*num2));
//
//
//		L_u.init_min_max(0.,1.);
//
//		if(a>0.){
//			L_u.add_limit_min(u_to_alpha(aoE/Snu,um,iDu));
//			L_u.add_limit_max(u_to_alpha(( (1.-ly*nu)-sqrt( pow_2(ly-nu)+pow_2(aoE*Snu) ) )/pow_2(Snu),um,iDu));
//
//			if(L_u.is_unvalid()) return false;
//		}
//
//		if(b>0.){
//			L_u.add_limit_max(u_to_alpha(_fthb(sign_of(nu-ly)*thb),um,iDu));
//
//			if(L_u.is_unvalid()) return false;
//		}
//
//		if(ex){
//			E_u_m.init(_fthb(-thb0),_fthb(-thb1));
//			E_u_p.init(_fthb(+thb0),_fthb(+thb1));
//
//			if(E_u_m.is_valid()) u_limit_to_alpha(E_u_m,um,iDu);
//			if(E_u_p.is_valid()) u_limit_to_alpha(E_u_p,um,iDu);
//
//			E_u_m.apply_to_limit(L_u);
//			E_u_p.apply_to_limit(L_u);
//
//			if(L_u.is_unvalid()) return false;
//		}
//
//		return true;
//	}


	inline double _integrand_nu(const double nu){
		double result(0);
		this->nu	= nu;
		Snu			= sqrt(1.-pow_2(nu));
		if(_calc_alphas()){
			result = f(L_u);
			if(ex){
				if(E_u_m.is_valid())	result -= f(E_u_m);
				if(E_u_p.is_valid())	result -= f(E_u_p);
			}
		}
		return result;
	}

	inline static double _gsl_integrand(double x, void* params){
		Two_P_PS* p = static_cast<Two_P_PS*>(params);
//		return p->_integrand_x(x);
		return p->_integrand_nu(x);
	}

	/***
	 * Return the value of nu, at which u_max(b) intercepts u+-
	 * -1 <= (--) <= (-+) <= (++) <= (+-) <= +1
	 */
	inline double _fnub_upm(const double sthb, const double su=-1.)const{
		const double a(lT*(sthb+ly));
		const double b(0.5*lM2+ly*(sthb+ly));
		const double c((sthb+ly)-0.5*lM2*sthb);

		const double D(pow_2(a)+pow_2(b)-pow_2(c));

		if(D<0.){
			return sign_of(sthb);
		}else{
			return (b*c-su*a*sqrt(D))/(pow_2(a)+pow_2(b));
		}
	}

	inline double _set0(){
		gsl_int.result	= 0.;
		gsl_int.error	= 0.;
		return 0.;
	}

	inline double _calc_one_region(const Variable_Limit& IR){
		if(IR.is_valid()){
			return gsl_int.integrate_a_to_b(IR.min,IR.max,acc_rel,acc_abs);
		}else{
			return 0.;
		}
	}

	inline double _calc_one_region(Variable_Limit& IR, Variable_Exclusion E1){
		E1.apply_to_limit(IR);
		if( E1.is_valid() && (IR.position_of(E1) == NS_Interval::I_I_inside) ){
			Variable_Limit IR2(E1.max,IR.max);
			IR.init_min_max(IR.min,E1.min);
			return _calc_one_region(IR)+_calc_one_region(IR2);
		}

		return _calc_one_region(IR);
	}

	inline double _calc_one_region(Variable_Limit& IR, const Variable_Exclusion& E1, Variable_Exclusion E2){
		E2.apply_to_limit(IR);
		if( E2.is_valid() && (IR.position_of(E2) == NS_Interval::I_I_inside) ){
			Variable_Limit IR2(E2.max,IR.max);
			IR.init_min_max(IR.min,E2.min);
			return _calc_one_region(IR,E1)+_calc_one_region(IR2,E1);
		}else{
			return _calc_one_region(IR,E1);
		}
	}

public:
	Two_P_PS();
	Two_P_PS(const Two_P_PS &RV);
	Two_P_PS& operator=(const Two_P_PS &RV);

	double init_and_return_min_mp(const double M);
	double init_and_return_min_mp(const double M, const double a, const double b);
	double init_and_return_min_mp(const double M, const double a, const double b, const double b0, const double b1);
	double init_and_return_min_mp(const double M, const OPS_Cuts &op);

	double update_qT_y(const double qT, const double y);

	ostream& into_stream(ostream& os)const;
};

ostream& operator<<(ostream &os, const Two_P_PS& PS);


class Two_P_PS_Interpolated: protected Two_P_PS {
protected:
	static const size_t _N_default;
	static const GSL_Interpolation_Types _Type_default;
	GSL_Interpolator gsl_interpolator;
	Variable_Limit y_interpolation_range;

public:
	double init_and_return_min_mp(const double M, const size_t N=_N_default, const GSL_Interpolation_Types Type=_Type_default);
	double init_and_return_min_mp(const double M, double a, const double b, const size_t N=_N_default, const GSL_Interpolation_Types Type=_Type_default);
	double init_and_return_min_mp(const double M, const double a, const double b, const double ex_b0, const double ex_b1, const size_t N=_N_default,  const GSL_Interpolation_Types Type=_Type_default);
	double init_and_return_min_mp(const double M, const OPS_Cuts &op_c, const size_t N=_N_default, const GSL_Interpolation_Types Type=_Type_default);

	const Variable_Limit& update_qT_get_ipol_range(const double qT, const Variable_Limit &L_y);

	inline const Variable_Limit& update_qT_get_ipol_range(const double qT, const double y_min, const double y_max){
		y_interpolation_range.init_min_max(y_min,y_max);
		return update_qT_get_ipol_range(qT,y_interpolation_range);
	}

	inline const Variable_Limit& update_qT_get_ipol_range(const double qT, const double y_max){
		return update_qT_get_ipol_range(qT,-y_max,y_max);
	}

	inline double get_interpolated_Jac(const double y)const{
		if(y_interpolation_range.position_of(y)==V_I_inside){
			if(cutted_PS())	return gsl_interpolator(fabs(y));
			else return 1.;
		}else return 0.;
	}

	inline double get_calculated_Jac(const double qT, const double y){
//		if(L_y.position_of(y)==V_I_inside){
		if(cutted_PS())	return Two_P_PS::update_qT_y(qT,y);
		else return 1.;
//		}else return 0.;
	}

	ostream& into_stream(ostream& os)const;
	ostream& into_stream(ostream& os, const size_t N)const;

	inline bool cutted_PS()const{ return cuts;}
};


}	//	NS_Kinematics
}	// CuTe



#endif
