#include "QDCS_Sigma0.h"
#include "param.h"
#include "Kinematics.h"

namespace CuTe {

namespace NS_QDCS {

using namespace NS_enums::NS_PROD;

using NS_CS_Params::CS_Parameter;

using namespace NS_Utils::NS_pow;

using std::cerr;
using std::complex;
using std::cout;
using std::endl;

class _Sigma_0{
public:
	constexpr static const double BARN = 0.389379323e9;
protected:

	double sigma_0;
	double dV;

	double	S;
	double	M;

	inline virtual void v_init(const CS_Parameter &p){
		std::cerr	<<"_Sigma_0::v_init() called -> exit(1)"<<endl;
		exit(1);
	}

	inline virtual void v_update(double as){
		std::cerr	<<"************************************************"<<endl
					<<"*** _Sigma_0::v_update(as) called -> exit(1) ***"<<endl
					<<"************************************************"<<endl;
		exit(1);
	}

public:
	inline void init(const CS_Parameter &p){
		sigma_0	= 0.;
		dV		= 0.;
		S		= p.S;
		M		= p.M;
		v_init(p);
	}

	inline double update(double qT, double as){
		v_update(as);
		dV=BARN*2.*qT;
		return sigma_0*dV;
	}

	inline _Sigma_0():sigma_0(0.),dV(0.),S(),M(){
	}

	inline _Sigma_0(const CS_Parameter &p)
	:sigma_0(0.),dV(0.)
	,S(p.S),M(p.M){
	}

	inline virtual ~_Sigma_0(){}
};

class Sigma_0_Z:public _Sigma_0{
protected:
	double ALPHA_EM;

	inline void v_init(const CS_Parameter &p){
		sigma_0	= 4.*PI2*ALPHA_EM/(N_c*S);
	}

	inline void v_update(double as){
	}

public:
	inline Sigma_0_Z(const CS_Parameter &p)
	:_Sigma_0(p)
	,ALPHA_EM(p.alpha_EM){
		v_init(p);
	}
};

class Sigma_0_gamma:public _Sigma_0{
protected:
	double AoM2;	//= (ALPHA_EM/M)^2;
	/* Note:
	 * AoM2 is in fact only ALPHA_EM^2 to generate
	 * M^2 d^2sigma / dqT dq^2 in pb/GeV
	 */

	inline void v_init(const CS_Parameter &p){
		sigma_0	= 4.*M_PI/(3.*N_c*S)*AoM2;
	}

	inline void v_update(double as){
	}
public:
	inline Sigma_0_gamma(const CS_Parameter &p)
	:_Sigma_0(p)
	,AoM2(pow_2(p.alpha_EM/*/M*/)){
		v_init(p);
	}
};

class Sigma_0_H:public _Sigma_0{
protected:
	double G_F;
	double M_t;

	inline complex<double> f_q(double x)const{
//		cout	<<"Sigma_0_H::f_q("<<x<<")"<<endl;
		if(x>=1.0) return(pow_2(asin(1.0/sqrt(x))));
		double f_q=sqrt(1.-x);
		f_q=(f_q+1.)/(f_q-1.);
		f_q=log(f_q);
		return -0.25*complex<double>(f_q,-M_PI);
	}

	inline virtual double A_q(double mq)const{
//		cout	<<"Sigma_0_H::A_q("<<mq<<")"<<endl;
		complex<double> A_q;
		double x;
		x=pow_2(2.0*mq/M);
		A_q=(f_q(x)*(1.0-x)+1.0)*1.5*x;
		A_q=A_q*conj(A_q);
		return A_q.real();
	}

	inline void v_init(const CS_Parameter &p){
//		cout	<<"Sigma_0_H::v_init(p)"<<endl;
	}

	inline void v_update(double as){
//		cout	<<"Sigma_0_H::v_update(as)"<<endl;
		sigma_0	= G_F*M_SQRT1_2*pow_2(4.*M_PI*as*M)/(288.*M_PI*S);

		if(M_t>0.)	sigma_0*=A_q(M_t);
	}
public:
	inline Sigma_0_H(const CS_Parameter &p)
	:_Sigma_0(p)
	,G_F(p.G_Fermi),M_t(p.M_t){
//		cout	<<"Sigma_0_H(p)"<<endl;
		v_init(p);
	}
};

class Sigma_0_H_RS:public Sigma_0_H{
protected:

	inline virtual double A_q(double mq)const{
//		cout	<<"Sigma_0_H_RS::A_q("<<mq<<")"<<endl;
		return 1.0;
	}

public:
	inline Sigma_0_H_RS(const CS_Parameter &p)
	:Sigma_0_H(p){
//		cout	<<"Sigma_0_H_RS(p)"<<endl;
	}
};



void Sigma_0::init(const CS_Parameter* p){
	_exit_s0();

	this->p	= p;

	prod	= p->prod;
	as		= &(p->as);

	_init_s0();
}


void Sigma_0::_exit_s0(){
	if(sigma0!=NULL){
		switch (prod) {
			case PROD_Higgs:
				delete static_cast<Sigma_0_H*>(sigma0);
				break;
			case PROD_H_RS:
				delete static_cast<Sigma_0_H_RS*>(sigma0);
				break;
			case PROD_DY:
				delete static_cast<Sigma_0_gamma*>(sigma0);
				break;
			case PROD_W:
			case PROD_Z:
				delete static_cast<Sigma_0_Z*>(sigma0);
				break;
			default:
				cout << "Sigma_0::init WRONG prod="<<p->prod<<"\t->exit"<<endl;
				exit(1);
				break;
		}
		sigma0 = NULL;
	}
}

void Sigma_0::_init_s0(){
	switch (prod) {
		case PROD_Higgs:
			sigma0 = new Sigma_0_H(*p);
			break;
		case PROD_H_RS:
			sigma0 = new Sigma_0_H_RS(*p);
			break;
		case PROD_DY:
			sigma0 = new Sigma_0_gamma(*p);
			break;
		case PROD_W:
		case PROD_Z:
			sigma0 = new Sigma_0_Z(*p);
			break;
		default:
			cout << "Sigma_0::init WRONG prod="<<p->prod<<"\t->exit"<<endl;
			exit(1);
			break;
	}
}

double Sigma_0::update_qT_mu(const double qT,const CS_Parameter* p){
	S0	= sigma0->update(qT,*as);
	return S0;
}

double Sigma_0::update_qT_mu(const Kinematics *Kin,const CS_Parameter* p){
	return update_qT_mu(Kin->qT,p);
}

}  // namespace NS_CROSSSECTION
}  // namespace CuTe
