#ifndef DD_INTEGRATOR_H
#define DD_INTEGRATOR_H

#include "differential_distributions.h"
#include "cuba_int.h"

namespace CuTe {

namespace CuTe_MC {

using namespace NS_diff_distris;
using NS_CUBA_integration::Cuba_Params;
using NS_CUBA_integration::cuba_integrand_t_vegas;
using Shared_Memory_Control::SM_Serial_ID;

using namespace NS_Utils::NS_pow;
using namespace NS_Utils::NS_min_max;



/*
 * Function to tranform the cuba array xx[dim] into proper coords
 * coords must have appropriate size before function call
 * must return the Jacobian
 */
typedef double(*Array_to_coords)(const double[], vector<double>& coords, void* params);

class DD_Integrator {
protected:
	SM_Serial_ID SID;
	Cuba_Params cuba;
	Differential_Distribution_Set DDset;
	Func_of_coords fx;
	Array_to_coords xx_to_vecx;
	void* params;

	vector<double> coords;

	inline static void _cuba_init(void* params, int &core){
		NS_Utils::DOUT dout("MC_params::CUBA_init_func");

		DD_Integrator* p = static_cast<DD_Integrator*>(params);
		p->SID.get_new_SID();

		dout()	<< &(p->SID)	<<"\t"<<"core="<<core	<<endl;

	}

	inline static void _cuba_exit(void* params, int &core){
		NS_Utils::DOUT dout("MC_params::CUBA_exit_func");

		DD_Integrator* p = static_cast<DD_Integrator*>(params);
		p->DDset.merge(*(p->SID));

		dout()	<< &(p->SID)	<<"\t"<<"core="<<core	<<endl;

	}

	inline void _integrand(const double xx[], double ff[],double weight, int iter){
		double vx = xx_to_vecx(xx,coords,params);
		*ff = fx(coords,params);
		DDset.add_value(*SID,coords,iter,*ff,weight,vx);
		(*ff) *= vx;
	}

	inline static int _cuba_integrand(
			const int *ndim, const double xx[]
			,const int *ncomp, double ff[], void *userdata
			,const int *nvec, const int* core
			,const double* weight, const int* iter) {

		DD_Integrator* p = static_cast<DD_Integrator*>(userdata);
		p->_integrand(xx,ff,(*weight),(*iter));
		return 0;
	}

public:

	inline DD_Integrator()
	:SID()
	,cuba()
	,DDset()
	,fx(NULL),xx_to_vecx(NULL)
	,params(NULL)
	,coords(){

	}

	inline void init(const SOPS_CuTe& OP, int dim,Func_of_coords fx, Array_to_coords xx_to_vecx, void* params, int ncomp=1){
		this->params = params;
		cuba.init(OP.CUBA,dim,ncomp);
		cuba.set(NS_CUBA_integration::CUBA_R_VEGAS,dim,ncomp);
		DDset.init(cuba.get_ncores(),params);
		this->cuba 			= cuba;
		this->fx			= fx;
		this->xx_to_vecx	= xx_to_vecx;
		coords.resize(dim);
	}

	inline void add_dist(const string& name, Func_of_coords x_of_coords, double min, double max, int N){
		DDset.add_dist(name,x_of_coords,min,max,N);
	}

	inline void cube(){
		cuba.set_in_ex(_cuba_init,_cuba_exit);
		cuba.cube((cuba_integrand_t_vegas)_cuba_integrand,this);
	}

	inline void to_File(const string& filename){
		DDset.to_File(filename);
	}

};

class Coordinates {
public:
	static const int DIM=4;
	enum e_coords{
		e_qT=0
		,e_y
		,e_nu
		,e_u
	};
protected:
	double sS;
	double S;

	double M;
	double M2;


	double qT_min,qT_max,qT_Delta,mp_max,mp;

	bool y_cut;
	double y_max_cut;
	double y_max;

	double q0,lM2,lT,ly;

	double nu_max,nu_min,nu_Delta,S_nu;

	double u_p,u_m,u_max,u_min,u_Delta;

	bool kT_cut;
	double a,a2,aoE;
	double nu_a_p,nu_a_m;
	double u_a_p,u_a_m;

	bool eta_cut;
	double b,thb;
	double u_b_p;

	vector<double> coords;
	double Jac;

	double tmp;

	inline double dv(){
		tmp	= coords[e_u];
		return 2.*M_1_PI*tmp/lM2*sqrt( (u_p*u_m)/((u_p-tmp)*(tmp-u_m)) );
	}

	inline bool _set_qT(const double xx_Tynu[4]){
		tmp	 = qT_min	+ qT_Delta*(*xx_Tynu);
		Jac	*= qT_Delta;

		return ( (Jac>0.) && set_qT(tmp) && _set_y(++xx_Tynu) );
	}

	inline bool _set_y(double y){
		coords[e_y]	= y;
		q0			= mp*cosh(y);
		ly			= tanh(y);

		tmp			= 1./q0;
		lM2			= M2*pow_2(tmp);
		lT			= coords[e_qT]*tmp;

		aoE			= a*tmp;

		if(eta_cut){
			nu_max	= thb;
		}else{
			nu_max	= 1.;
		}
		nu_min	= -nu_max;

		if(kT_cut){
			nu_a_p	= pow_2(2.*a*q0)*ly;
			tmp		= 1./( pow_4(mp)+nu_a_p*ly );
			nu_a_p	*= tmp;
			tmp		*= pow_3(mp)*sqrt(mp*mp-4.*a2);
			nu_a_m	= nu_a_p-tmp;
			nu_a_p	= nu_a_p+tmp;

			nu_max	= min_of(nu_max,nu_a_p);
			nu_min	= max_of(nu_min,nu_a_m);
		}

		nu_Delta	= nu_max	- nu_min;

		return (nu_Delta>0.);
	}

	inline bool _set_y(const double xx_ynu[3]){
		tmp	 = ((*xx_ynu)-0.5)*2.*y_max;
		Jac	*=	2.*y_max;

		return ( _set_y(tmp) && _set_nu(++xx_ynu) );
	}

	inline bool _set_nu(double nu){
		coords[e_nu]= nu;

		S_nu	= sqrt(1.-nu*nu);

		u_p	= 0.5*lM2/((1.-ly*nu) - lT*S_nu);
		u_m	= 0.5*lM2/((1.-ly*nu) + lT*S_nu);

		u_max	= u_p;
		u_min	= u_m;

		if(kT_cut){
			u_a_p	= ( 1.-ly*nu-sqrt(pow_2(ly-nu)+pow_2(aoE*S_nu)) )/pow_2(S_nu);
			u_a_m	= aoE/S_nu;

			u_max	= min_of(u_max,u_a_p);
			u_min	= max_of(u_min,u_a_m);
		}

		u_Delta	= u_max-u_min;
		if(u_Delta<0.) return false;

		if(eta_cut){
			tmp		= sign_of(ly-nu);
			u_b_p	= ( thb -tmp*ly )/( thb - tmp*nu );

			u_max	= min_of(u_max,u_b_p);

			u_Delta	= u_max-u_min;
		}

		return u_Delta>0.;
	}

	inline bool _set_nu(const double xx_nu[2]){
		tmp	 = nu_min + nu_Delta*(*xx_nu);
		Jac	*=	nu_Delta;

		return ( _set_nu(tmp) && _set_u(++xx_nu) );
	}

	inline bool _set_u(const double xx_u[1]){
		double u	= u_min + u_Delta*(*xx_u);
		coords[e_u]= u;

		Jac	*=	u_Delta;
		return (Jac>0.);
	}


	inline double _set_coords(const double xx_Tynu[], vector<double>& coords){
		Jac	= 1.;

		if(_set_qT(xx_Tynu)){
			coords	= this->coords;
		}else{
			Jac	= 0.;
		}
		return Jac*dv();
	}

public:

	inline Coordinates()
	:sS(),S()
	,M(),M2()
	,qT_min(),qT_max(),qT_Delta(),mp_max(),mp()
	,y_cut(),y_max_cut(),y_max()
	,q0(),lM2(),ly(),lT()
	,nu_max(),nu_min(),nu_Delta(),S_nu()
	,u_p(),u_m(),u_max(),u_min(),u_Delta()
	,kT_cut(),a(),a2(),aoE()
	,nu_a_p(),nu_a_m()
	,u_a_p(),u_a_m()
	,eta_cut(),b(),thb()
	,u_b_p()
	,coords(DIM,0.)
	,Jac()
	,tmp(){

	}

	inline void init(double sS, double M){
		this->sS	= sS;
		this->M		= M;

		S	= sS*sS;
		M2	= M*M;


		mp_max	= (S+M2)/(2.*sS);
		set_1_qT_cut(-1.,-1.);
		set_2_y_cut(-1.);
		set_3_kT_cut(-1.);
		set_4_eta_cut(-1.);
	}

	inline void set_1_qT_cut(double qT_min,double qT_max){
		if(qT_min<0.)	qT_min	= 0.;
		if(qT_max<=0.)	qT_max	= (S-M2)/(2.*sS);

		this->qT_min	= qT_min;
		this->qT_max	= qT_max;
		qT_Delta		= qT_max-qT_min;
	}

	inline void set_2_y_cut(double y_max){
		if(y_max<=0.){
			y_cut		= false;
		}else{
			y_cut		= true;
			y_max_cut	= y_max;
		}
	}

	inline void set_3_kT_cut(double a){
		if(a<=0.){
			kT_cut		= false;
		}else{
			kT_cut		= true;
			this->a		= a;
			a2			= a*a;
			qT_min		= max_of(qT_min,4.*a2-M2);
			qT_Delta	= qT_max-qT_min;
		}
	}

	inline void set_4_eta_cut(double b){
		if(b<=0.){
			eta_cut		= false;
		}else{
			eta_cut		= true;
			this->b		= b;
			thb			= tanh(b);
			if(y_cut){
				y_max_cut	= min_of(y_max_cut,b);
			}else{
				set_2_y_cut(b);
			}
		}
	}

	inline bool set_qT(double qT){
		coords[e_qT]	= tmp;

		mp		= sqrt(M2+pow_2(tmp));
		y_max	= acosh(mp_max/mp);
		if(y_cut)	y_max = min_of(y_max,y_max_cut);

		return (y_max>0.);
	}

	inline bool set_qT_y(double qT, double y){
		return ( set_qT(qT)&&_set_y(y) );
	}

	static inline double set_coords(const double xx[], vector<double>& coords, void* params){
		Coordinates* C	= static_cast<Coordinates*>(params);
		return C->_set_coords(xx,coords);
	}

};


namespace Test_Integrator {

	static inline double test_u(const vector<double>& coords, void* params){
		return 0.5*(coords[0]+coords[1]);
	}
	static inline double test_v(const vector<double>& coords, void* params){
		return 0.5*(coords[0]-coords[1]);
	}
	static inline double test_x(const vector<double>& coords, void* params){
		return coords[0];
	}
	static inline double test_y(const vector<double>& coords, void* params){
		return coords[1];
	}

	static inline double test_f(const vector<double>& coords, void* params){
		//return pow_2(coords[0]-10.*coords[1]);
		//return pow_3(coords[0]+coords[1]);
		return M_PI*sin(M_PI_2*coords[0]);
	}

	inline double test_set_coords(const double xx[], vector<double>& coords, void* params){
		coords[0] = xx[0];
		coords[1] = xx[1];
		return 1.;
	}

	inline void test(){
		string filename("MCtest");
		SOPS_CuTe OP(filename);
		Cuba_Params cuba;
		cuba.init(OP.CUBA,2,1);
		int N = 20;
		int dim =2;

		DD_Integrator DDI;
		DDI.init(OP,dim,test_f,test_set_coords,NULL);
		DDI.add_dist("u",test_u, 0., 1.,N);
		DDI.add_dist("v",test_v,-0.5,0.5,N);
		DDI.add_dist("x",test_x, 0., 1.,N);
		DDI.add_dist("y",test_y, 0., 1.,N);

		DDI.cube();

		DDI.to_File(filename);

	}



}  // namespace Test_Integrator


}  // namespace CuTe_MC



}  // namespace CuTe

#endif
