#ifndef SM_CONTROL_H
#define SM_CONTROL_H


#include <sys/ipc.h>
#include <sys/shm.h>
#include <boost/interprocess/sync/interprocess_mutex.hpp>

namespace CuTe {
namespace Shared_Memory_Control {

using std::cout;
using std::endl;
using std::string;
using std::ostream;
typedef unsigned long int ulint;

class __SM_base{
protected:
	int SMID;

	__SM_base():SMID(-1){}
	__SM_base(int smid):SMID(smid){}

public:
	inline int status(const string name)const{
		shmid_ds info;
		int status = shmctl(SMID, IPC_STAT, &info);
		cout	<< name << ".status=" << status << endl
				<< "\tSMID=" << SMID << endl
				<< "\tshm_segsz=" << info.shm_segsz << "\t(Byte)" << endl
				<< "\tshm_atime=" << info.shm_atime << "\t(Last attach time)" << endl
				<< "\tshm_dtime=" << info.shm_dtime << "\t(Last detach time)" << endl
				<< "\tshm_ctime=" << info.shm_ctime << "\t(Last change time)" << endl
				<< "\tshm_cpid="  << info.shm_cpid	<< "\t(PID of creator)" << endl
				<< "\tshm_lpid="  << info.shm_lpid	<< "\t(PID of last at-/detach)" << endl
				<< "\tshm_nattch="<< info.shm_nattch<< "\t(current # of attachs)"<< endl;
		return status;
	}

	inline ostream& status(ostream& os, const string name)const{
		shmid_ds info;
		int status = shmctl(SMID, IPC_STAT, &info);
		os	<< name << ".status=" << status << endl
				<< "\tSMID=" << SMID << endl
				<< "\tshm_segsz=" << info.shm_segsz << "\t(Byte)" << endl
				<< "\tshm_atime=" << info.shm_atime << "\t(Last attach time)" << endl
				<< "\tshm_dtime=" << info.shm_dtime << "\t(Last detach time)" << endl
				<< "\tshm_ctime=" << info.shm_ctime << "\t(Last change time)" << endl
				<< "\tshm_cpid="  << info.shm_cpid	<< "\t(PID of creator)" << endl
				<< "\tshm_lpid="  << info.shm_lpid	<< "\t(PID of last at-/detach)" << endl
				<< "\tshm_nattch="<< info.shm_nattch<< "\t(current # of attachs)"<< endl;
		return os;
	}

	inline int status(shmid_ds* info)const{
		return shmctl(SMID, IPC_STAT, info);
	}
};

template < typename T >
class SM_Variable: public __SM_base{
protected:
	T* val;

public:
	inline SM_Variable<T>()
	:__SM_base(shmget(IPC_PRIVATE, sizeof(T*), IPC_CREAT | 0666))
	,val(static_cast<T*>(shmat(SMID, 0, 0))){
		new (val) T;
	}

	inline SM_Variable<T>(const SM_Variable<T>& other)
	:__SM_base(shmget(IPC_PRIVATE, sizeof(T*), IPC_CREAT | 0666))
	,val(static_cast<T*>(shmat(SMID, 0, 0))){
		new (val) T(*other);
//		(*val) = (*other);
	}

	inline ~SM_Variable(){
		shmdt(val);
		shmctl(SMID, IPC_RMID, 0);
	}

	inline T& operator*(){return *val;}
	inline const T& operator*()const{return *val;}
};

template <>
class SM_Variable<boost::interprocess::interprocess_mutex>: public __SM_base{
protected:
	typedef boost::interprocess::interprocess_mutex mutex_T;
	mutex_T* val;
public:
	inline SM_Variable<mutex_T>()
	:__SM_base(shmget(IPC_PRIVATE, sizeof(mutex_T*), IPC_CREAT | 0666))
	,val(static_cast<mutex_T*>(shmat(SMID, 0, 0))){
		new (val) mutex_T;
	}

	inline SM_Variable<mutex_T>(const SM_Variable<mutex_T>& other)
	:__SM_base(shmget(IPC_PRIVATE, sizeof(mutex_T*), IPC_CREAT | 0666))
	,val(static_cast<mutex_T*>(shmat(SMID, 0, 0))){
		new (val) mutex_T;
	}

	inline ~SM_Variable(){
		shmdt(val);
		shmctl(SMID, IPC_RMID, 0);
	}

	inline void lock(){val->lock();}
	inline bool try_lock(){return val->try_lock();}
	inline void unlock(){val->unlock();}
};

typedef SM_Variable<boost::interprocess::interprocess_mutex> Mutex;



template < typename T >
class SM_lockable_Variable{
protected:
	Mutex mutex;
	SM_Variable<T> var;

	SM_lockable_Variable(const SM_lockable_Variable&);
	SM_lockable_Variable& operator=(const SM_lockable_Variable&);

public:
	inline SM_lockable_Variable()
	:mutex(),var(){

	}

	inline ~SM_lockable_Variable(){

	}

	inline T& operator*(){return *var;}
	inline const T& operator*()const{return *var;}

	inline void 	lock(){return mutex.lock();}
	inline bool try_lock(){return mutex.try_lock();}
	inline void   unlock(){return mutex.unlock();}

};

class SM_Serial_ID{
protected:
	SM_Variable<int> counter;
	mutable Mutex mutex;
	int SID;
public:

	inline SM_Serial_ID()
	:SID(0){
		(*counter)=0;
	}

	inline int get_new_SID(){
		mutex.lock();
		SID=(++(*counter));
		mutex.unlock();
		return SID;
	}

	inline int get_counter()const{
		mutex.lock();
		int get_counter=(*counter);
		mutex.unlock();
		return get_counter;
	}

	inline ~SM_Serial_ID(){
		mutex.lock();
		--(*counter);
		mutex.unlock();
	}

	inline std::ostream& status(std::ostream& os)const{
		mutex.lock();
		os	<<"SM_Serial_ID.status:"	<<"\t"<<"SID="<<SID	<<"\t"<<"counter="<<*counter;
		mutex.unlock();
		return os;
	}

//	inline int operator*()const{return SID;}
	inline const int operator*()const{return SID;}

};

inline std::ostream& operator<< (std::ostream& os,const SM_Serial_ID* SMSID){
	return SMSID->status(os);
}


template < typename T >
class SM_Array: public __SM_base{
protected:
	ulint N;
	T* val;
	bool attached;
public:

	inline int init(ulint size){
		if(attached) clear();
		N=(size);
		SMID = shmget(IPC_PRIVATE, N*sizeof(T*), IPC_CREAT | 0666);
		val = static_cast<T*>(shmat(SMID, 0, 0));
		attached=true;
		return SMID;
	}

	inline int init(ulint size, const T& t){
		int erg = init(size);
		for(T* here = begin(); here < end(); ++here){
			(*here) = t;
		}
		return erg;
	}

	inline void clear(){
		shmdt(val);
		shmctl(SMID, IPC_RMID, 0);
		attached = false;
	}

	inline SM_Array<T>()
	:__SM_base()
	,N(0)
	,val(NULL)
	,attached(false){
	}

	inline SM_Array<T>(ulint size)
	:__SM_base(shmget(IPC_PRIVATE, size*sizeof(T*), IPC_CREAT | 0666))
	,N(size)
	,val(static_cast<T*>(shmat(SMID, 0, 0)))
	,attached(true){
	}

	inline T* begin()const{ return val;}
	inline T* end()const{ return &(val[N]);}

	inline SM_Array<T>(const SM_Array<T>& other)
	:__SM_base(shmget(IPC_PRIVATE, other.size()*sizeof(T*), IPC_CREAT | 0666))
	,N(other.size())
	,val(static_cast<T*>(shmat(SMID, 0, 0)))
	,attached(true){
		T* there = other.begin();
		for(T* here = begin(); here < end(); ++here){
			(*here) = (*(there++));
		}
	}

	inline ~SM_Array(){
		clear();
	}


	inline T &operator [] (ulint const i){ return val[i];}
	inline T const &operator [] (ulint const i)const {return val[i];	}

	inline ulint size()const{return N;}
};

}  // namespace Shared_Memory_Control



}  // namespace CuTe

#endif
